/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */


/*
 * Routines to send TCP reset for the SF firewall packet filter
 * 
 * $Id: sf_tcp.c,v 1.0 1990/07/25 14:37:44 robby Rel $
 *
 */

#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS 1
#endif
#include <linux/modversions.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/icmp.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/ip.h>
#include <linux/malloc.h>
#include <linux/timer.h>
#include "sf_kernel.h"
#include "sf_rc.h"
#include "sf_global.h"
#include "sf_filter.h"
#include "sf_control.h"
#include "sf_tcp.h"
#include <net/checksum.h>
#include <asm/checksum.h>

#define PRINTIP(a) printk(KERN_ALERT "%lu.%lu.%lu.%lu", ntohl(a)>>24&0xFF,\
                          ntohl(a)>>16&0xFF, ntohl(a)>>8&0xFF, ntohl(a)&0xFF)

/*   I think I've reactivated the send_tcp(), although I haven't tested this
 *   _thoroughly_ (only on 100 MBit ethernet, needs to be tested for other
 *   network-hardware)... mail me if you encounter problems/have suggestions
 *   (sky@real-linux.de)                1999/10/13 Florian Heinz
 */

void send_tcp(int rst, struct device *dev, struct iphdr *ip, struct tcphdr *tcp)
{
  struct sk_buff *skb;
  struct rtable *rt;
  struct tcphdr *tcp2;
  __u32 saddr, daddr;
  int len, st;

  if (dev->flags & IFF_LOOPBACK)	/* Aiiieee :-) */
    return;

  len = sizeof(struct iphdr) + sizeof(struct tcphdr);

  if (tcp->rst)			/* Cannot reply to reset */
    return;

#ifdef SF_DEBUG_TCP
  printk("sf: tcp: send_tcp()\n");
#endif

  skb = dev_alloc_skb(len);
  if (skb == NULL)
    return;

  /* Initializing sk-buffer */
  skb->nh.iph = (struct iphdr *) skb_put(skb, 20);
  skb->h.th = (struct tcphdr *) skb_put(skb, sizeof(struct tcphdr));
  rt = (struct rtable *) skb->dst;
  skb->dev = dev;
  skb->sk = NULL;
  skb->pkt_type = PACKET_OTHERHOST;
  skb->csum = 0;
  skb->protocol = __constant_htons(ETH_P_IP);

  /* Generating IP-header */
  memcpy(skb->nh.iph, ip, sizeof(struct iphdr));

  saddr = ip->daddr;
  daddr = ip->saddr;
  skb->nh.iph->saddr = saddr;
  skb->nh.iph->daddr = daddr;
  skb->nh.iph->ttl = 255;
  skb->nh.iph->tot_len = htons(len);
  skb->nh.iph->frag_off = 0;
  skb->nh.iph->ihl = 5;
  skb->nh.iph->check = 0;
  skb->nh.iph->check = ip_fast_csum((char *) skb->nh.iph, skb->nh.iph->ihl);
   
  /* Calculating route... */
  if ((st = ip_route_output(&rt, daddr, 0, 0, 0)) != 0) {
#ifdef SF_DEBUG_TCP
    printk(KERN_ALERT "sf: send_tcp: Calculating routes failed! errno: %d",
           st);
#endif
    kfree_skb(skb);
    return;
  }
  skb->dst = &rt->u.dst;

  /* generating TCP header */
  tcp2 = skb->h.th;
  memset(tcp2, 0, sizeof(struct tcphdr));
  tcp2->source = tcp->dest;
  tcp2->dest   = tcp->source;
  tcp2->doff   = sizeof(struct tcphdr) >> 2;

  if (tcp->ack)
    tcp2->seq = tcp->ack_seq;
  else {
    tcp2->ack = 1;
    if (!tcp->syn)
      tcp2->ack_seq = tcp->seq;
    else
      tcp2->ack_seq = htonl(ntohl(tcp->seq) + 1);
  }
  if (rst)
    tcp2->rst = 1;
  
  tcp2->check = csum_tcpudp_magic(saddr, daddr, sizeof(struct tcphdr),
	   IPPROTO_TCP, csum_partial((char *) tcp2, sizeof(struct tcphdr), 0));
  
#ifdef SF_DEBUG_TCP
  printk("sf: tcp: %s seq=%08lx, ack_seq=%08lx\n", rst ? "RST" : "ACK",
         ntohl(tcp2->seq), ntohl(tcp2->ack_seq));
#endif
   

  /* And send the packet... */
  ip_finish_output(skb);
}
